﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.SymbolicExpressionTreeEncoding {
  /// <summary>
  /// A base class for operators creating symbolic expression trees.
  /// </summary>
  [Item("TracingSymbolicExpressionTreeCreator", "A base class for operators creating symbolic expression trees.")]
  [StorableClass]
  public abstract class TracingSymbolicExpressionTreeCreator : SymbolicExpressionTreeOperator, ISymbolicExpressionTreeCreator {
    private const string SymbolicExpressionTreeParameterName = "SymbolicExpressionTree";
    private const string GlobalTreeCacheParameterName = "GlobalTreeCache";


    #region Parameter Properties
    public ILookupParameter<ISymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (ILookupParameter<ISymbolicExpressionTree>)Parameters[SymbolicExpressionTreeParameterName]; }
    }
    #endregion

    #region Properties
    public ISymbolicExpressionTree SymbolicExpressionTree {
      get { return SymbolicExpressionTreeParameter.ActualValue; }
      set { SymbolicExpressionTreeParameter.ActualValue = value; }
    }

    #endregion
    [StorableConstructor]
    protected TracingSymbolicExpressionTreeCreator(bool deserializing) : base(deserializing) { }
    protected TracingSymbolicExpressionTreeCreator(TracingSymbolicExpressionTreeCreator original, Cloner cloner) : base(original, cloner) { }
    protected TracingSymbolicExpressionTreeCreator()
      : base() {
      Parameters.Add(new LookupParameter<ISymbolicExpressionTree>(SymbolicExpressionTreeParameterName, "The symbolic expression tree that should be created."));
      Parameters.Add(new LookupParameter<ItemList<ISymbolicExpressionTree>>(GlobalTreeCacheParameterName, "The global cache for storing the trees"));
      //globalScope.Variables.Add(new Variable(GlobalOperatorCacheParameterName, new ItemList<ISymbolicExpressionTree>()));
    }

    public override IOperation Apply() {
      SymbolicExpressionTree = Create(Random);
      // add tree to the global tree cache
      AddTreeToGlobalCache();

      return base.Apply();
    }

    private void AddTreeToGlobalCache() {
      var globalScope = ExecutionContext.Scope;
      while (globalScope.Parent != null) globalScope = globalScope.Parent;

      Variable globalTreeCache;
      if (!globalScope.Variables.ContainsKey(GlobalTreeCacheParameterName)) {
        globalTreeCache = new Variable(GlobalTreeCacheParameterName, new ItemList<ISymbolicExpressionTree>());
        globalScope.Variables.Add(globalTreeCache);
      } else
        globalTreeCache = (Variable)globalScope.Variables[GlobalTreeCacheParameterName];
      var trees = (ItemList<ISymbolicExpressionTree>)globalTreeCache.Value;
      trees.Add(SymbolicExpressionTree);
    }

    protected abstract ISymbolicExpressionTree Create(IRandom random);
    public abstract ISymbolicExpressionTree CreateTree(IRandom random, ISymbolicExpressionGrammar grammar, int maxTreeLength, int maxTreeDepth);
  }
}
